
from collections import defaultdict
from pathlib import Path
import fnmatch
import logging
import re

import numpy as np

from extra_data import by_id
from extra_data.read_machinery import select_train_ids

from exdf import DataFile


exdf_filename_pattern = re.compile(
    r'([A-Z]+)-R(\d{4})-(\w+)-S(\d{5}).h5')


def get_source_origin(sourcedata):
    # TODO: Move to EXtra-data.

    try:
        files = sourcedata.source_files
    except AttributeError:
        files = sourcedata.files

    data_category = set()
    run_numbers = set()
    aggregators = set()

    for f in files:
        m = exdf_filename_pattern.match(Path(f.filename).name)

        if m:
            data_category.add(m[1])
            run_numbers.add(int(m[2]))
            aggregators.add(m[3])

    result_sets = [data_category, run_numbers, aggregators]

    if any([len(x) != 1 for x in result_sets]):
        raise ValueError('none or multiple sets of traits recognized '
                         'in source filenames, inconsistent input dataset?')

    return data_category.pop(), run_numbers.pop(), aggregators.pop()


def sourcedata_drop_empty_trains(sd):
    # TODO: Move to EXtra-data.

    if sd._is_control:
        train_sel = sd[next(iter(sd.keys()))].drop_empty_trains().train_ids
    else:
        sample_keys = dict(zip(
            [key[:key.find('.')] for key in sd.keys()], sd.keys()))

        train_sel = np.zeros(0, dtype=np.uint64)

        for key in sample_keys.values():
            train_sel = np.union1d(
                train_sel, sd[key].drop_empty_trains().train_ids)

    return sd.select_trains(by_id[train_sel])


def sourcedata_data_counts(sd):
    # TODO: Move to EXtra-data

    if sd._is_control:
        return sd[next(iter(sd.keys()))].data_counts(labelled=False)
    else:
        sample_keys = dict(zip(
            [key[:key.find('.')] for key in sd.keys()], sd.keys()))

        data_counts = {
            prefix: sd[key].data_counts(labelled=False)
            for prefix, key in sample_keys.items()
        }
        return data_counts


class ReductionSequence:
    # A sequence of reduction methods.

    log = logging.getLogger('exdf.data_reduction.ReductionSequence')

    def __init__(self, data, methods):
        self._data = data
        self._methods = methods

        self._instructions = sum(
            [m._instructions for m in methods.values()], [])

        self._sources = sorted(data.all_sources)
        self._touched_sources = set()

        # Only populated if trains/keys are selected/removed for sources.
        self._custom_keys = {}  # source -> set(<keys>)
        self._custom_trains = {}  # source -> list(<trains>)
        self._partial_copies = {}  # (source, key) -> (region, chunks)

    def _filter_ops(self, op):
        return [args[1:] for args in self._instructions if args[0] == op]

    def link(self, output_scope):
        """Link together multiple methods into a sequence."""

        for source_glob, in self._filter_ops('removeSources'):
            for source in fnmatch.filter(self._sources, source_glob):
                self._touched_sources.add(source)
                self._sources.remove(source)

        for source_glob, key_glob in self._filter_ops('removeKeys'):
            for source in fnmatch.filter(self._sources, source_glob):
                self._touched_sources.add(source)

                keys = self._custom_keys.setdefault(
                    source, set(self._data[source].keys()))

                for key in fnmatch.filter(keys, key_glob):
                    keys.remove(key)

        for source_glob, train_sel in self._filter_ops('selectTrains'):
            for source in fnmatch.filter(self._sources, source_glob):
                self._touched_sources.add(source)
                train_ids = self._custom_trains.setdefault(
                    source, list(self._data.train_ids))

                self._custom_trains[source] = select_train_ids(
                    train_ids, train_sel)

        for source_glob, key, region, chunking in self._filter_ops(
            'partialCopy'
        ):
            for source in fnmatch.filter(self._sources, source_glob):
                self._touched_sources.add(source)
                self._partial_copies[(source, key)] = (region, chunking)

        if output_scope == 'sources':
            self._sources = sorted(
                self._touched_sources.intersection(self._sources))

        elif output_scope == 'aggregators':
            touched_aggregators = {get_source_origin(self._data[source])[2]
                                   for source in self._touched_sources}

            self._sources = sorted(
                {source for source in self._sources
                 if (get_source_origin(self._data[source])[2]
                     in touched_aggregators)})

        if not self._sources:
            self.log.warn('Linking reduction methods yielded empty source '
                          'selection')

    def write_collection(self, output_path, sequence_len=512):
        if not self._sources:
            self.log.error('No sources to write to collection.')
            return

        outp_data = self._data.select([(s, '*') for s in self._sources])

        # An origin is a pair of (storage_class, aggregator)
        origin_sources = defaultdict(list)

        run_numbers = set()

        # Collect all origins (combination of data category and
        # aggregator) and the sources they contain.
        for source in self._sources:
            data_category, run_no, aggregator = get_source_origin(
                outp_data[source])  # TODO: Use support in EXtra-data
            origin_sources[(data_category, aggregator)].append(source)
            run_numbers.add(run_no)

        if len(run_numbers) != 1:
            self.log.error('Could not infer run number from filenames')
            return

        run_no = run_numbers.pop()

        for (data_category, aggregator), sources in origin_sources.items():
            # Build source & key selection for this origin.
            origin_sel = {
                s: self._custom_keys[s] if s in self._custom_keys else set()
                for s in sources}

            # Select output data down to what's in this origin both in
            # terms of sources and trains (via require_any).
            origin_data = outp_data \
                .select(origin_sel, require_any=True)

            # Switch to representation of SourceData objects for
            # per-source tracking of trains.
            origin_sources = {source: origin_data[source]
                              for source in origin_data.all_sources}

            # Apply custom train selections.
            for source, train_sel in self._custom_trains.items():
                if source in origin_sources:
                    origin_sources[source] = origin_sources[source] \
                        .select_trains(by_id[train_sel])

            # Find the union of trains across all sources as total
            # trains for this origin.
            origin_train_ids = np.zeros(0, dtype=np.uint64)
            for sd in origin_sources.values():
                origin_train_ids = np.union1d(
                    origin_train_ids,
                    sourcedata_drop_empty_trains(sd).train_ids)

            num_trains = len(origin_train_ids)
            num_sequences = int(np.ceil(num_trains / sequence_len))

            self.log.info(
                f'{data_category}-{aggregator}: {len(origin_sources)} sources '
                f'with {num_trains} trains over {num_sequences} sequences')

            for seq_no in range(num_sequences):
                # Slice out the train IDs for this sequence.
                seq_train_ids = origin_train_ids[
                    (seq_no * sequence_len):((seq_no + 1) * sequence_len)]

                # Select origin data down to what's in this file.
                seq_sources = {source: sd.select_trains(by_id[seq_train_ids])
                               for source, sd in origin_sources.items()}

                seq_path = self._write_sequence(
                    output_path, aggregator, run_no, seq_no, data_category,
                    seq_sources, seq_train_ids)
                self.log.debug(
                    f'{seq_path.stem}: {len(seq_sources)} sources with '
                    f'{len(seq_train_ids)} trains')

    def write_voview(self, output_path):
        raise NotImplementedError('voview output layout')

    def write_collapsed(self, output_path, sequence_len):
        raise NotImplementedError('collapsed collection output layout')

    def _write_sequence(self, folder, aggregator, run, sequence, data_category,
                        sources, train_ids=None):
        """Write collection of SourceData to file."""

        if train_ids is None:
            pass  # Figure out as union of source_train_ids

        sd_example = next(iter(sources.values()))

        # Build sources and indices.
        control_indices = {}
        instrument_indices = {}

        for source, sd in sources.items():
            if sd._is_control:
                control_indices[sd.source] = sourcedata_data_counts(sd)
            else:
                instrument_indices[sd.source] = sourcedata_data_counts(sd)

        with DataFile.from_details(folder, aggregator, run, sequence,
                                   prefix=data_category, mode='w') as f:
            output_path = Path(f.filename)

            # TODO: Handle old files without extended METADATA?
            # TODO: Handle timestamps, origin
            # TODO: Attributes
            # TODO: Chunking of run/control data?

            # Create METADATA section.
            f.create_metadata(
                like=sd_example,
                control_sources=control_indices.keys(),
                instrument_channels=[
                    f'{source}/{channel}'
                    for source, channels in instrument_indices.items()
                    for channel in channels.keys()])

            # Create INDEX section.
            f.create_index(train_ids)

            for source, counts in control_indices.items():
                control_src = f.create_control_source(source)
                control_src.create_index(len(train_ids), per_train=True)

            for source, channel_counts in instrument_indices.items():
                instrument_src = f.create_instrument_source(source)
                instrument_src.create_index(**channel_counts)

            # Create CONTROL and RUN sections.
            for source in control_indices.keys():
                exdf_source = f.source[source]
                sd = sources[source]

                run_data_leafs = {}

                for path, value in sd.run_values().items():
                    key = path[:path.rfind('.')]
                    leaf = path[path.rfind('.')+1:]

                    run_data_leafs.setdefault(key, {})[leaf] = value

                for key, leafs in run_data_leafs.items():
                    exdf_source.create_run_key(key, **leafs)

                for key in sd.keys(False):
                    exdf_source.create_key(key,
                                           sd[f'{key}.value'].ndarray(),
                                           sd[f'{key}.timestamp'].ndarray(),
                                           run_entry=False)

            # Create INSTRUMENT datasets.
            for source in instrument_indices.keys():
                exdf_source = f.source[source]
                sd = sources[source]

                for key in sd.keys():
                    kd = sd[key]

                    shape = (kd.data_counts(labelled=False).sum(),
                             *kd.entry_shape)
                    try:
                        _, chunks = self._partial_copies[source, key]
                    except KeyError:
                        chunks = kd.files[0].file[kd.hdf5_data_path].chunks

                    exdf_source.create_key(
                        key, shape=shape, maxshape=(None,) + shape[1:],
                        chunks=chunks, dtype=kd.dtype)

            # Copy INSTRUMENT data.
            for source in instrument_indices.keys():
                exdf_source = f.source[source]
                sd = sources[source]

                for key in sd.keys():
                    # TODO: Copy by chunk / file if too large

                    data = sd[key].ndarray()

                    try:
                        region, _ = self._partial_copies[source, key]
                    except KeyError:
                        exdf_source.key[key][:] = data
                    else:
                        full_region = (np.s_[:], *region)
                        exdf_source.key[key][full_region] = data[full_region]

        return output_path
